/*
    v0.01 - original
    v0.02 - support for 13.26, requires exTap v0.12
          - display series/repeats from highlighted epg entry - press recall
    v0.03 - lists channel number
          - firmware 13.40, exTap_v0.15
    v0.04 - merged in dx's fixes - overwritng of crids due to now/next
    v0.05 - added scroll bar to window
    v0.06 - fixed key events
    v0.07 - support for v14.02, built in series link
    v0.08 - 14.06
    v0.08a - 14.08

crids do not seem to be present on table:43 (now/next, present mux)
how to retain the schedule?

*/
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "eit.h"

#define ORIGINAL 1

#define ID_USERTAP 0xbdb01e08
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("crid v0.08");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("a tap to extract Series Link information");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#include "TSRCommander.inc"

extern void TSRCommanderConfigDialog();
extern bool TSRCommanderExitTAP();
extern int is_tsr_commander_running(void);

//----------------------------------------------------------------------------
//
//main
//
extern type_menu main_menu;

extern int TAP_Main(void);
extern dword TAP_EventHandler(word event, dword param1, dword param2);
extern int main_menu_handler(type_menu *menu, dword param1);
extern int key_handler(dword param1);
extern int key_handler2(dword param1);

extern int init(void);
extern int close(void);

extern void init_crc(void);
extern dword crc8(dword crc, byte *data, int len);

type_eventtable *get_current_event(int service_id);
type_eventtable *get_event(int service_id, int event_id);
type_eventtable *get_event_from_crid(byte *crid_str);
int dump_crid(type_eventtable *current_event);
dword epg_crid(dword param1);
//----------------------------------------------------------------------------
int debug = 0;

int eit_enabled = 0;
int eit_hook_id = -2;
int apply_eit_hook();
int remove_eit_hook();
dword eit_hook(dword param1, dword param2);
dword eit_filter(byte *from);

extern int genre_enabled;
extern int genre_hook_id;
extern int apply_genre_hook();
extern int remove_genre_hook();
extern dword genre_hook(dword param1, dword param2);

dword service_id_filter(byte *buf);

word crid_rgn = 0;
int crid_showing = 0;
int current_crid = 0;
int current_epg = 0;
TYPE_Window crid_window;

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    TAP_ExitNormal();
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    if (fw_tap_table <= 0) return 0;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}

//----------------------------------------------------------------------------
//
//main
//
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {4, "xxx",                 },
        {0, "Current Event Status",},
        {1, "Current Timer Status",},
        {2, "Enable Serial Debug", },
        {3, "Exit Tap",            },
    },
};

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;

    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    ret = exTAP_Init();
    if (ret < 0) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return 0;
    }
    if ((fw_event_table == 0) || (fw_event_table_size == 0) || (fw_max_events == 0) ||
        (fw_timer_table == 0) || (fw_timer_table_size == 0) || (fw_max_timers == 0)) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return 0;
    }

    init();
    TSRCommanderInit(0, TRUE);

    return 1;                 //0=application mode, 1=tsr mode
}

int init(void)
{
    debug = 0;
    eit_enabled = 0;
    genre_enabled = 0;
    main_menu.visible = 0;
    init_crc();

    crid_rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
    if ((short)crid_rgn == -1) {
        TAP_Print("\nShowMessage_y(rgn=%x)", crid_rgn);
        crid_rgn = 0;
//        return 0;
    }

    if ((fw_sysid==456)&&(_appl_version>=0x1400)) {
    }
    else {
        apply_eit_hook();
    }
    apply_genre_hook();

    return 0;
}

int close(void)
{
    sprintf(strLog, "Closing %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    remove_eit_hook();
    remove_genre_hook();

    DeleteMenu(&main_menu);
    TAP_EnterNormal();
    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (crid_rgn) TAP_Osd_Delete(crid_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();

//    TAP_Print("\nClosing %s\n",__tap_program_name__);
    TAP_Exit();
    return 0;
}

#ifndef ORIGINAL

#define SECOND_IN_TICKS     100

typedef enum
{
    RUNNING_STATUS_Undefined,
    RUNNING_STATUS_NotRunning,
    RUNNING_STATUS_StartsSoon,
    RUNNING_STATUS_Pausing,
    RUNNING_STATUS_Running,
    RUNNING_STATUS_Reserved
} TYPE_RunningStatus;

static dword CurrentTime (void)
{
    byte hour, min, sec;
    word mjd;

    TAP_GetTime (&mjd, &hour, &min, &sec);
    return (mjd << 16) + (hour << 8) + min;
}

static void RefreshEIT (byte SvcType, word SvcIndex)
{
    TYPE_TapEvent *eventInfoTable = NULL;
    int eventNum = 0;

    // get EIT tables
    if ((eventInfoTable = TAP_GetEvent (SvcType, SvcIndex, &eventNum)) && (eventNum > 1))
    {
        dword time = CurrentTime ();
        int i;
        // process events and update the running status
        for (i=0; ((i < eventNum) && (eventInfoTable[i].startTime <= time)); i++)
        {
            byte runningStatus = RUNNING_STATUS_NotRunning;
            // if this the current event ?
            if ((eventInfoTable[i].startTime <= time) && (time < eventInfoTable[i].endTime))
            {
                runningStatus = RUNNING_STATUS_Running;
            }
            // if status isn't what we expect update it
            if (eventInfoTable[i].runningStatus != runningStatus)
            {
                TAP_EPG_UpdateEvent (SvcType, SvcIndex, eventInfoTable[i].evtId, runningStatus);
            }
        }
        TAP_MemFree (eventInfoTable);
    }
}


static void UpdateWork (void)
{
    static byte SvcType = 0;
    static word SvcIndex = 0;
    int NumTvSvc, NumRadioSvc;
    TYPE_TapEvent *eventInfo = NULL;

    TAP_Channel_GetTotalNum (&NumTvSvc, &NumRadioSvc);
    if (SvcType == 0)
    {
        if (++SvcIndex >= NumTvSvc)
        {
            SvcType = 1;
            SvcIndex = 0;
        }
    }
    else
    {
        if (++SvcIndex >= NumRadioSvc)
        {
            SvcType = 0;
            SvcIndex = 0;
        }
    }
    RefreshEIT (SvcType, SvcIndex);
}
#endif

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    TYPE_TapChInfo channel_info;
    int channel_type, channel_number;
    int ret;
    dword state, subState;
    static dword tick = 0;

    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                    TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
#ifndef ORIGINAL
            if ((TAP_GetTick () - tick) > SECOND_IN_TICKS)
            {
                UpdateWork ();
                tick = TAP_GetTick ();
            }
#endif
            TSRCommanderWork();
            break;

        case EVT_KEY:
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
            TAP_GetState(&state, &subState);
//            TAP_Print("\nEVT_KEY, state=%x, subState=%x", state, subState);
//printf("\ncurrent_crid = %x", current_crid);
//printf("\ncrid_showing = %x", crid_showing);

            if ((state == STATE_Epg) && (subState == SUBSTATE_Epg)) {
                //in epg mun
                if (crid_showing) {
                    if ((param1 == RKEY_Exit) || (param1 == RKEY_Recall) || (param1 == RKEY_Sat)) {
                        TAP_Win_Delete(&crid_window);
                        if (fw_epg_refresh) *fw_epg_refresh = 0x18;   //set state == epg:epg
                        crid_showing = 0;
//                        param1 = 0;
                        return param1;
                    }
                }
                else
                if (param1 == RKEY_Recall) {
                    current_epg = 1;
                    param1 = epg_crid(param1);
                }
                else
                if (param1 == RKEY_Sat) {   //swap pip
                    current_epg = 1;
                    dump_timers();
                }
                else
                if (param1 == RKEY_Exit) {
                    TAP_Win_Delete(&crid_window);
                    crid_showing = 0;
//                    param1 = 0;
                    return param1;
                }
            }
            else
            if (main_menu.visible) {
                param1 = main_menu_handler(&main_menu, param1);
            }
            else
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                param1 = key_handler(param1);
            }

            if (current_epg) {
                param1 = key_handler2(param1);
                if (param1 == RKEY_Exit) {
                    TAP_Win_Delete(&crid_window);
                    if (fw_epg_refresh) *fw_epg_refresh = 0x18;   //set state == epg:epg
                    current_crid = 0;
                    crid_showing = 0;
                    current_epg = 0;
                    param1 = 0;
                    return param1;
                }
            }

            if (current_crid) {
                param1 = key_handler2(param1);
                if (param1 == RKEY_Exit) {
                    TAP_Win_Delete(&crid_window);
                    TAP_EnterNormal();
                    current_crid = 0;
                    crid_showing = 0;
                    param1 = 0;
                    return param1;
                }
            }
            else
            if (crid_showing) {
                param1 = key_handler2(param1);
                if (param1 == RKEY_Exit) {
                    TAP_Win_Delete(&crid_window);
                    TAP_EnterNormal();
//                    if (fw_epg_refresh) *fw_epg_refresh = 0x18;   //set state == epg:epg
                    crid_showing = 0;
                    return param1;
                }
            }
            break;
        case EVT_HOOK:
            if (param1 == eit_hook_id) param1 = eit_hook(param1, param2);
            if (param1 == genre_hook_id) param1 = genre_hook(param1, param2);
            break;
        default:
//            TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int i;
    int num_events;
    int ret;
    int tvRadio;
    int chNum;
    int nTvSvc;
    int nRadioSvc;
    TYPE_TapChInfo chInfo;
    type_eventtable *current_event;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            TAP_EnterNormal();
            switch (selection) {
                case 0:
                    TAP_Channel_GetCurrent(&tvRadio, &chNum);
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, chNum, &chInfo);
                    current_event = get_current_event(chInfo.svcId);
                    if (current_event) {
                        current_crid = 1;
                        TAP_ExitNormal();
                        ret = dump_crid(current_event);
                    }
                    return 0;
                case 1:
                    current_crid = 1;
                    TAP_ExitNormal();
                    ret = dump_timers();
                    return 0;
                case 2:
                    debug ^= 1;
                    if (debug) {
                        sprintf(strLog, "Enabling Serial Debug");
                        strcpy(menu->item[3].str, "Disable Serial Debug");
                    }
                    else {
                        sprintf(strLog, "Disabling Serial Debug");
                        strcpy(menu->item[3].str, "Enable Serial Debug");
                    }
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    return 0;
                case 3:
                    close();
                    return 0;
            }
            break;
       case RKEY_Recall:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return param1;
        case RKEY_Exit:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return 0;
        default:
            TAP_Win_Action(&menu->win, param1);
            return 0;
    }
    return param1;
}


int key_handler(dword param1)
{
    switch(param1) {
       //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                TAP_ExitNormal();
                ShowMenu(&main_menu);
                return 0;
            }
    }
    return param1;
}

int key_handler2(dword param1)
{
    int selection;

    switch(param1) {
        case RKEY_Up:
             TAP_Win_Action(&crid_window, RKEY_Up);
             return 0;
             break;
        case RKEY_Down:
            TAP_Win_Action(&crid_window, RKEY_Down);
            return 0;
            break;
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&crid_window);
            return 0;
            break;
    }
    return param1;
}
//----------------------------------------------------------------------------
//#define EPG_TABLE
#ifdef EPG_TABLE
typedef struct{
    int valid;
    int table_id;
//    int section_length;
    int table_id_extension;
    int version_number;
//    int current_next_indicator;
    int section_number;
//    int last_section_number;
//    int event_id;
//    int count;
//    int len;
} type_table;

#define MAX_TABLE_ID    0x2000
int num_tables;
type_table *table;
#endif

//-----------------------------------------------------------------------------
//
//eit stuff
//
//hook the process_eit() function
//this code just filters by service_id
//allow through all now/next (table_id=0x4e)
//only allow through enabled schedule (table_id=50/51)
//could add max# days per channels
//should never allow if servicve_id does not exists - e.g when 4tv data
//
type_hook eit_hook_patch[] = {
// sysid, firmware, enable, address,    length, original_data
//    {456, 0x1209,   1,      0x8014ed60, 8,      {0x27, 0xbd, 0xff, 0x28,  0xaf, 0xb0, 0x00, 0xb8, }, },  //fw_process_eit
//    {456, 0x1225,   1,      0x8014fc2c, 8,      {0x27, 0xbd, 0xff, 0x28,  0xaf, 0xb0, 0x00, 0xb8, }, },  //fw_process_eit
    {456, 0x1288,   1,      0x80155194, 8,      {0x27, 0xbd, 0xff, 0x18,  0xaf, 0xb0, 0x00, 0xc8, }, },  //fw_process_eit
    {0},
};
type_hook eit_hook_patch2[] = {
// sysid, firmware, enable, address,    length, original_data
    {456, 0x1402,   1,      0x8015d140, 8,      {0x27, 0xbd, 0xfe, 0xe8,  0xaf, 0xb0, 0x00, 0xf8, }, },  //fw_process_eit
    {0},
};

byte *to;

int apply_eit_hook(void)
{
    int i;

    if (fw_process_eit == 0) {
        eit_enabled = 0;
        TAP_Print("\nERROR: apply_eit_hook(), firmware version %04x not supported", _appl_version);
        return -1;
    }
    //fix the address;
    eit_hook_patch[0].sysid = fw_sysid;
    eit_hook_patch[0].revision = _appl_version;
    eit_hook_patch[0].enable = 1;
    eit_hook_patch[0].addr = (dword)fw_process_eit;

    eit_hook_patch2[0].sysid = fw_sysid;
    eit_hook_patch2[0].revision = _appl_version;
    eit_hook_patch2[0].enable = 1;
    eit_hook_patch2[0].addr = (dword)fw_process_eit;

#ifdef EPG_TABLE
    num_tables = 0;
    table = (type_table *)malloc(sizeof(type_table) * MAX_TABLE_ID);
    if (table == 0) {
        sprintf(strLog, "ERROR: apply_eit_hook(), malloc failed, table=%x, len=%x", table, sizeof(type_table) * MAX_TABLE_ID);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }
    memset(table, 0, sizeof(type_table) * MAX_TABLE_ID);
#endif

    to = (byte *)malloc(0x1000);
    if (to == 0) {
#ifdef EPG_TABLE
        free(table);
#endif
        sprintf(strLog, "ERROR: apply_eit_hook(), malloc failed, to = %x, len=%x", to, 0x1000);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }

    eit_hook_id = exTAP_ApplyHook(&eit_hook_patch[0]);
    if (eit_hook_id >= 0) {
        eit_enabled = 1;
        return 0;
    }

    eit_hook_id = exTAP_ApplyHook(&eit_hook_patch2[0]);
    if (eit_hook_id >= 0) {
        eit_enabled = 1;
        return 0;
    }

    eit_enabled = 0;
#ifdef EPG_TABLE
    free(table);
#endif
    free(to);
    sprintf(strLog, "ERROR: apply_eit_hook(), firmware version %04x not supported", _appl_version);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
    return -1;
}

int remove_eit_hook(void)
{
    int ret;

    if (eit_hook_id >=0) {
        ret = exTAP_RemoveHook(eit_hook_id);
        if (ret == 0) {
            eit_hook_id = -1;
        }
    }
    free(to);
#ifdef EPG_TABLE
    free(table);
#endif
    eit_enabled = 0;
    return 0;
}

dword eit_hook(dword param1, dword param2)
{
    type_hook_registers *reg;
    int ret;

//TAP_Print("\nEVT_HOOK_eit, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
    if (is_firmware(reg->a0)) {
        reg->a0 = eit_filter((byte *)(reg->a0));
    }
    return param1;
}

//----------------------------------------------------------------------------
dword eit_filter(byte *from)
{
    dword crc;
    byte table_id;
    word section_length;
    word table_id_extension;
    byte version_number;
    byte current_next_indicator;
    byte section_number;
    byte last_section_number;
    word transport_stream_id;
    word original_network_id;

#ifdef EPG_TABLE
    int found;
    type_table *t;
#endif
    int i;
    int to_len;

    //do crc check
    section_length = 3 + (((from[1]<<8) + from[2]) & 0xfff);
    crc = crc32(from, section_length);
    if (crc) {
        if (debug & 1) {
            TAP_Print("\nWARNING: *** CRC ERROR ***, section_length=%x", section_length);
            dump((dword)&from[0], 0x40);
        }
        return (dword)from;
    }

    table_id               = from[0];
    section_length         = 3 + (((from[1]<<8) + from[2]) & 0xfff);
    table_id_extension     = ((from[3]<<8) + from[4]) & 0xffff;
    version_number         = (from[5]>>1) & 0x1f;
    current_next_indicator = (from[5]>>0) & 0x1;
    section_number         = from[6];
    last_section_number    = from[7];
    transport_stream_id    = ((from[8]<<8) + from[9]) & 0xffff;         //eit only
    original_network_id    = ((from[10]<<8) + from[11]) & 0xffff;       //eit only

//    TAP_Print("\nEIT:  0012:%02x:%04x %3d/%3d: %08x:%-3x", table_id, table_id_extension, section_number, last_section_number, from, section_length);

#ifdef EPG_TABLE
    //there may be multiple sections
    //each table has the same table_id, table_id_extension
    //use last_section_number to determine the number of sections per table
    //process each section independantly
    //but keep a count

    found = -1;
    for (i=0; i<MAX_TABLE_ID; i++) {
        if ((table[i].table_id == table_id) && (table[i].table_id_extension == table_id_extension) && (table[i].section_number == section_number)) {
            t = (type_table *)&table[i];
            if (debug & 2) TAP_Print("\nEIT-%02x:%04x %3d/%3d: matched table, %x", table_id, table_id_extension, section_number, last_section_number, i);
//            if (debug & 1) TAP_Print("m");
            found = i;
            goto cont1;
        }
    }
cont1:
    if (found < 0) {
        t = (type_table *)&table[num_tables];
        if (debug & 2) TAP_Print("\nEIT-%02x:%04x %3d/%3d: newtable, %x", table_id, table_id_extension, section_number, last_section_number, num_tables);
        if (debug & 1) TAP_Print("n");
        if (num_tables >= MAX_TABLE_ID) {
            sprintf(strLog,"ERROR: EIT-%02x:%04x %3d/%3d: out of tables - %x", num_tables, table_id, table_id_extension, section_number, last_section_number);
            log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
            //clear out + start again ...
            num_tables = 0;
            memset(table, 0, sizeof(type_table) * MAX_TABLE_ID);
            return (dword)from;
        }
        num_tables++;
    }

    t->table_id = table_id;
    t->table_id_extension = table_id_extension;
    t->section_number = section_number;
//    t->last_section_number = last_section_number;
//    t->section_length = section_length;
//    t->len = len;
//    t->count ++;


    //have we already processed this section
    //has it changed
    if (t->valid) {
        if (t->version_number == version_number) {
            if (debug & 2) TAP_Print("\nEIT-%02x:%04x %3d/%3d: already got section", table_id, table_id_extension, section_number, last_section_number);
            return (dword)from;
        }
        else {
            if (debug & 1) TAP_Print("\nEIT-%02x:%04x %3d/%3d: new version, was %2x, now %2x", table_id, table_id_extension, section_number, last_section_number, t->version_number, version_number);
            t->valid = 0;
        }
    }

    t->valid = 1;
    t->version_number = version_number;
#endif

    //reject unless eit
    switch (table_id) {
        case 0x4e:
        case 0x4f:
#ifdef ORIGINAL
            //allow all now_next
            break;
#else
            // disallow all now_next (doesn't contain crid's)
            return (dword)from;
#endif
        case 0x50:
        case 0x51:
        case 0x60:
        case 0x61:
            break;
        default:
            sprintf(strLog,"WARNING: eit_filter(), invalid table_id = %02x", from[0]);
            log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
            return (dword)from;
            break;
    }

    //now process eit section
    convert_event_information_section(from, to, &to_len);

    return (dword)&to[0];
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//*
//* The FCS-32 generator polynomial: x**0 + x**1 + x**2 + x**4 + x**5
//*                      + x**7 + x**8 + x**10 + x**11 + x**12 + x**16
//*                      + x**22 + x**23 + x**26 + x**32.
//*
#define P32             0xedb88320  // the polymonial
#define CRC_POLYNOMIAL  0x04c11db7
//*
//* FCS lookup table as calculated by the table generator.
//*
static dword fcstab8[0x100];

static dword bitrevN(int n, dword data)
{
    int i;
    dword result;

    result = 0;
    for (i=0; i<n; i++) {
        result += ((data >> (n-1-i)) & 1) << (i);
    }
    return result;
}

void init_crc(void)
{
    int b;
    int b2;
    dword v;
    int i;

    for (b=0; b<(1<<8); b++) {
        b2 = bitrevN(8,b);
        v = (dword)b;
        for(i=0; i<8; i++) {
            v = (v >> 1) ^ ((v & 1) ? P32 : 0);
        }

        fcstab8[b2] = bitrevN(32,v);
    }
}

// 32 bit crc, 8 bit data, 8 bits at a time
// lookup table = 2^8=256 words
dword crc8(dword crc, byte *data, int len)
{
    int i;
    int index;
    dword dataout32;

    for(i=0; i<len; i++) {
        index = ((crc>>24) ^ (data[i])) & 0xff;
        dataout32 = fcstab8[index];
        crc = (crc << 8) ^ dataout32;
    }

    return crc;
}

//----------------------------------------------------------------------------
void dump_event(type_eventtable *fw_et, int event_index, char *str)
{
    int i,j;
    char event_name[0x100];
    char *description;
    word year;
    byte month, day, weekDay;
    dword utc, local;
    word start_date;
    byte start_hour, start_min;
    word end_date;
    byte end_hour, end_min;
    int nTvSvc, nRadioSvc;
    TYPE_TapChInfo chInfo;
    char str1[0x100];


    if (fw_et->char00 & 0xc0) {

        utc = (fw_et->start_date << 16) + (fw_et->start_hour<<8) + fw_et->start_min;
        local = exTAP_UTCtoLocalTime(utc);
        start_date = (local>>16)&0xffff;
        start_hour = (local>>8)&0xff;
        start_min = (local>>0)&0xff;
        TAP_ExtractMjd(start_date, &year, &month, &day, &weekDay);

        sprintf(str1, " ");
        strcpy(str, str1);
        if (start_date) {
            sprintf(str1, "%02d/%02d/%04d", day, month, year);
            strcat(str, str1);
        }
        sprintf(str1, ", %02d:%02d", start_hour, start_min);
        strcat(str, str1);

        utc = (fw_et->end_date << 16) + (fw_et->end_hour<<8) + fw_et->end_min;
        local = exTAP_UTCtoLocalTime(utc);
        end_date = (local>>16)&0xffff;
        end_hour = (local>>8)&0xff;
        end_min = (local>>0)&0xff;
        TAP_ExtractMjd(end_date, &year, &month, &day, &weekDay);

        sprintf(str1, "-%02d:%02d", end_hour, end_min);
        strcat(str, str1);


        TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
        for (i=0; i<nTvSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &chInfo);
            if (fw_et->service_id == chInfo.svcId) {
                goto found;
            }
        }
        for (i=0; i<nRadioSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &chInfo);
            if (fw_et->service_id == chInfo.svcId) {
                goto found;
            }
        }
        goto not_found;

found:
//        sprintf(str1, " Ch. %d", chInfo.logicalChNum);
        sprintf(str1, " %s -", chInfo.chName);
        strcat(str, str1);

not_found:
        memcpy(event_name, fw_et->event_name, fw_et->event_name_length);
        event_name[fw_et->event_name_length] = '\0';
        sprintf(str1, " %s", event_name);
        strcat(str, str1);
    }
}

void dump_timer(type_timer_block *fw_tb, int timer_index, char *str)
{
    int i,j;
    char event_name[0x100];
    char *description;
    word year;
    byte month, day, weekDay;
    word start_date;
    byte start_hour, start_min;
    word end_date;
    byte end_hour, end_min;
    int nTvSvc, nRadioSvc;
    TYPE_TapChInfo chInfo;
    char str1[0x100];
    type_eventtable *fw_et;


    if (fw_tb->flags & 0xe0) {
        start_date = (fw_tb->startTime>>16)&0xffff;
        start_hour = (fw_tb->startTime>>8)&0xff;
        start_min = (fw_tb->startTime>>0)&0xff;
        TAP_ExtractMjd(start_date, &year, &month, &day, &weekDay);

//        sprintf(str1, " ");
        sprintf(str1, "%d. ", timer_index);
        strcpy(str, str1);
        if (start_date) {
            sprintf(str1, "%02d/%02d/%04d", day, month, year);
            strcat(str, str1);
        }
        sprintf(str1, ", %02d:%02d", start_hour, start_min);
        strcat(str, str1);

        end_date = (fw_tb->endTime>>16)&0xffff;
        end_hour = (fw_tb->endTime>>8)&0xff;
        end_min = (fw_tb->endTime>>0)&0xff;
        TAP_ExtractMjd(end_date, &year, &month, &day, &weekDay);

        sprintf(str1, "-%02d:%02d", end_hour, end_min);
        strcat(str, str1);


        TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
        for (i=0; i<nTvSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &chInfo);
            if (fw_tb->service_id == chInfo.svcId) {
                goto found;
            }
        }
        for (i=0; i<nRadioSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &chInfo);
            if (fw_tb->service_id == chInfo.svcId) {
                goto found;
            }
        }
        goto not_found;

found:
//        sprintf(str1, " Ch. %d", chInfo.logicalChNum);
        sprintf(str1, " %s -", chInfo.chName);
        strcat(str, str1);

not_found:
        //get the event info for this entry
        if (fw_tb->event_id) {
            fw_et = get_event(fw_tb->service_id, fw_tb->event_id);
            if (is_firmware(fw_et)) {
                memcpy(event_name, fw_et->event_name, fw_et->event_name_length);
                event_name[fw_et->event_name_length] = '\0';
                sprintf(str1, " %s", event_name);
                strcat(str, str1);
            }
        }


//        //get the event info for this original entry
//        if ((fw_tb->crid_valid == 1) && (fw_tb->event_id3)) {
//            fw_et = get_event(fw_tb->service_id, fw_tb->event_id3);
//            if (is_firmware(fw_et)) {
//                memcpy(event_name, fw_et->event_name, fw_et->event_name_length);
//                event_name[fw_et->event_name_length] = '\0';
//                sprintf(str1, "\n     %s", event_name);
//                strcat(str, str1);
//            }
//        }
    }
}
//----------------------------------------------------------------------------
//
//scan epg from the end
//check for multiple running on a channel
//remove the earliest
//
type_eventtable *get_current_event(int service_id)
{
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;
    dword utc, local, now, start, end;
    word start_date;
    word mjd;
    byte hour, min, sec;

    //goto the start of the chain
    event_index = 0;
    k = 0;
    while (k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

        if ((fw_et->char00 & 0xc0) == 0) {
            //find a valid entry
            event_index++;
        }
        else {
            //is this the start
            if (fw_et->previous == 0xffff) {
                break;
            }
            event_index = fw_et->previous;
        }
        k++;
    }
    if (k == fw_max_events) {
        //epg empty
        TAP_Print("\nepg empty");
        return 0;
    }


    //get current time - we want this in utc
    exTAP_GetUTCTime(&mjd, &hour, &min);
    now = (mjd << 16) + (hour<<8) + (min);

    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //match service_id
        if (fw_et->service_id == service_id) {

            //scan up until we get the current entry
            start = (fw_et->start_date << 16) + (fw_et->start_hour<<8) + fw_et->start_min;
            end = (fw_et->end_date << 16) + (fw_et->end_hour<<8) + fw_et->end_min;
            if (now >= end) {
                //ignore it - event over
            }
            else
            if (now < start) {
                //event not started; gone too far!
                //abort
                break;
            }
            else {
                return fw_et;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

//    TAP_Print("\ncannot find current event");
    return 0;
}

type_eventtable *get_event(int service_id, int event_id)
{
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;

//printf("\nservice_id=%x, event_id=%x", service_id, event_id);
    //goto the start of the chain
    event_index = 0;
    k = 0;
    while (k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

        if ((fw_et->char00 & 0xc0) == 0) {
            //find a valid entry
            event_index++;
        }
        else {
            //is this the start
            if (fw_et->previous == 0xffff) {
                break;
            }
            event_index = fw_et->previous;
        }
        k++;
    }
    if (k == fw_max_events) {
        //epg empty
        TAP_Print("\nepg empty");
        return 0;
    }

//printf("\nevent_index=%x", event_index);

    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //match service_id
        if (fw_et->service_id == service_id) {

            //match event_id
            if (fw_et->event_id == event_id) {

                return fw_et;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

//    TAP_Print("\ncannot find current event");
    return 0;
}


type_eventtable *get_event_from_crid(byte *crid_str)
{
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;

//printf("\nservice_id=%x, event_id=%x", service_id, event_id);
    //goto the start of the chain
    event_index = 0;
    k = 0;
    while (k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

        if ((fw_et->char00 & 0xc0) == 0) {
            //find a valid entry
            event_index++;
        }
        else {
            //is this the start
            if (fw_et->previous == 0xffff) {
                break;
            }
            event_index = fw_et->previous;
        }
        k++;
    }
    if (k == fw_max_events) {
        //epg empty
        TAP_Print("\nepg empty");
        return 0;
    }

//printf("\nevent_index=%x", event_index);

    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //match crid_str
        if (fw_et->crid32_strlen) {
            if (strncmp(fw_et->crid32_str, crid_str, fw_et->crid32_strlen) == 0) {

//printf("\nstrcmp %d, %s, %s", fw_et->crid32_strlen, fw_et->crid32_str, crid_str);
                return fw_et;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

//    TAP_Print("\ncannot find current event");
    return 0;
}

int dump_crid(type_eventtable *current_event)
{
    dword state, subState;
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;
    dword utc, local, now, start, end;
    word start_date;
    word mjd;
    byte hour, min, sec;
    word mjd1;
    byte hour1, min1;

    char event_name[0x100];
    byte num_extended;
    word extended_length;
    byte *extended_event_name;
    int offset, ooffset;

    int found_current = 0;
    char *crid31 = NULL;
    char *crid32 = NULL;
    char *crid33 = NULL;
    int series_count = 0;
    int repeat_count = 0;
    int recommend_count = 0;
    TYPE_TapChInfo chInfo;
    char str[0x100];
    char str1[0x100];

    dword dwRgn, dwCol, dwRow, dwWidth, dwHeight, dwMaxCol;

    //delete any old ones
    if (crid_showing) {
        TAP_Win_Delete(&crid_window);
//        TAP_EnterNormal();
//        TAP_ExitNormal();
        if (current_crid == 0) {
            if (fw_epg_refresh) *fw_epg_refresh = 0x18;   //set state == epg:epg
        }
        crid_showing = 0;
    }

    dwWidth = 500;
    dwCol = ((720 - dwWidth) / 2) - 25;
    dwRow = 100;
    dwHeight = 300;
    dwMaxCol = dwCol + 18 + dwWidth + 18;

    TAP_Win_SetDefaultColor(&crid_window);
    TAP_Win_Create(&crid_window, crid_rgn, dwCol, dwRow, dwWidth, dwHeight, TRUE, TRUE);
    TAP_Win_SetTitle(&crid_window, "SeriesLink", 1, FNT_Size_1622);
    dwRow = 150;
    TAP_GetState(&state, &subState);
    if (current_crid == 0) {
          if (fw_epg_refresh) *fw_epg_refresh = 0x3c;
    }
    crid_showing = 1;

    sprintf(str1, "\n");
    strcpy(str, str1);

    //!!!the order of the internal type_eventtable structure is different for v12.88!!!
    if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
        num_extended = current_event->event.fw13.num_extended;
        extended_length = current_event->event.fw13.extended_length;
        extended_event_name = current_event->event.fw13.extended_event_name;
    }
    else {
        num_extended = current_event->event.fw.num_extended;
        extended_length = current_event->event.fw.extended_length;
        extended_event_name = current_event->event.fw.extended_event_name;
    }

    //if we have a crid, remember it + continue scanning for identical ones
    if ((fw_sysid==456)&&(_appl_version>=0x1400)) {
        //check for "crid=31"
        if (current_event->crid31_strlen && current_event->crid31_str) {
            crid31 = current_event->crid31_str;
            sprintf(str1,"RepeatId:  %9s", current_event->crid31_str);
            strcpy(str, str1);
            dump_event(current_event, 0, str1);
            strcat(str, str1);
            if (debug & 1) TAP_Print("\n%s", str);
        }
        //check for "crid=32"
        if (current_event->crid32_strlen && current_event->crid32_str) {
            crid32 = current_event->crid32_str;
            sprintf(str1,"SeriesLink:%9s", current_event->crid32_str);
            strcpy(str, str1);
            dump_event(current_event, 0, str1);
            strcat(str, str1);
            if (debug & 1) TAP_Print("\n%s", str);
        }
        //check for "crid=33"
        if (current_event->crid33_strlen && current_event->crid33_str) {
            crid33 = current_event->crid33_str;
            sprintf(str1,"Recommend: %9s", current_event->crid33_str);
            strcpy(str, str1);
            dump_event(current_event, 0, str1);
            strcat(str, str1);
            if (debug & 1) TAP_Print("\n%s", str);
        }
    }
    else {
        offset = 0;
        for (i=0; i<num_extended; i++) {
            ooffset = offset;
            offset += strlen(&extended_event_name[offset]);
            offset ++;

            //check for "crid=31"
            if (strncmp(&extended_event_name[ooffset], "crid=31", 7) == 0) {
                if (extended_event_name[offset]) {
                    crid31 = &extended_event_name[offset];
                    sprintf(str1,"RepeatId:  %9s", &extended_event_name[offset]);
                    strcpy(str, str1);
                    dump_event(current_event, 0, str1);
                    strcat(str, str1);
                    if (debug & 1) TAP_Print("\n%s", str);
                }
            }
            //check for "crid=32"
            if (strncmp(&extended_event_name[ooffset], "crid=32", 7) == 0) {
                if (extended_event_name[offset]) {
                    crid32 = &extended_event_name[offset];
                    sprintf(str1,"SeriesLink:%9s", &extended_event_name[offset]);
                    strcpy(str, str1);
                    dump_event(current_event, 0, str1);
                    strcat(str, str1);
                    if (debug & 1) TAP_Print("\n%s", str);
                }
            }
            //check for "crid=33"
            if (strncmp(&extended_event_name[ooffset], "crid=33", 7) == 0) {
                if (extended_event_name[offset]) {
                    crid33 = &extended_event_name[offset];
                    sprintf(str1,"Recommend: %9s", &extended_event_name[offset]);
                    strcpy(str, str1);
                    dump_event(current_event, 0, str1);
                    strcat(str, str1);
                    if (debug & 1) TAP_Print("\n%s", str);
                }
            }

            offset += strlen(&extended_event_name[offset]);
            offset ++;
        }
    }


    //goto the start of the chain
    event_index = 0;
    k = 0;
    while (k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

        if ((fw_et->char00 & 0xc0) == 0) {
            //find a valid entry
            event_index++;
        }
        else {
            //is this the start
            if (fw_et->previous == 0xffff) {
                break;
            }
            event_index = fw_et->previous;
        }
        k++;
    }
    if (k == fw_max_events) {
        //epg empty
        TAP_Print("\nepg empty");
        return -1;
    }


    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //!!!the order of the internal type_eventtable structure is different for v12.88!!!
        if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
            num_extended = fw_et->event.fw13.num_extended;
            extended_length = fw_et->event.fw13.extended_length;
            extended_event_name = fw_et->event.fw13.extended_event_name;
        }
        else {
            num_extended = fw_et->event.fw.num_extended;
            extended_length = fw_et->event.fw.extended_length;
            extended_event_name = fw_et->event.fw.extended_event_name;
        }

        if ((fw_sysid==456)&&(_appl_version>=0x1400)) {
            //check for "crid=32"
            if (fw_et->crid32_strlen && fw_et->crid32_str) {
//                 TAP_Print("\n32:%s", fw_et->crid32_str);
                if (crid32) {
                    if (crid32[0]) {
                        if (strcmp(fw_et->crid32_str, crid32) == 0) {
                            if (series_count == 0) {
//                              TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "SeriesLink", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                TAP_Win_AddItem(&crid_window, "SeriesLink");
                                dwRow += 20;
                            }
                            series_count ++;
//                             sprintf(str1,"%9s", fw_et->crid32_str);
//                             strcpy(str, str1);
                            dump_event(fw_et, j, str1);
                            strcpy(str, str1);
//                          TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                            TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                            TAP_Win_AddItem(&crid_window, str);
                            dwRow += 20;


//                //get the crid for this original entry
//                if (fw_et->crid32_strlen) {
//                    sprintf(str, "\n     %s [R:%s]", fw_et->crid32_str, (fw_et->crid31_strlen)?fw_et->crid31_str:"");
//
////                  TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
//                    TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
//                    TAP_Win_AddItem(&crid_window, str);
//                    dwRow += 20;
//                }

                            if (debug & 1) TAP_Print("\n%s", str);
                        }
                    }
                }
            }
        }
        else {
            offset = 0;
            for (i=0; i<num_extended; i++) {
                ooffset = offset;
                offset += strlen(&extended_event_name[offset]);
                offset ++;


                //check for "crid=32"
                if (strncmp(&extended_event_name[ooffset], "crid=32", 7) == 0) {
                    if (extended_event_name[offset]) {
    //                    TAP_Print("\n32:%s", &extended_event_name[offset]);
                        if (crid32) {
                            if (crid32[0]) {
                                if (strcmp(&extended_event_name[offset], crid32) == 0) {
                                    if (series_count == 0) {
//                                      TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "SeriesLink", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                        TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                        TAP_Win_AddItem(&crid_window, "SeriesLink");
                                        dwRow += 20;
                                    }
                                    series_count ++;
    //                                sprintf(str1,"%9s", &extended_event_name[offset]);
    //                                strcpy(str, str1);
                                    dump_event(fw_et, j, str1);
                                    strcpy(str, str1);
//                                  TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                    TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                    TAP_Win_AddItem(&crid_window, str);
                                    dwRow += 20;

                                    if (debug & 1) TAP_Print("\n%s", str);
                                }
                            }
                        }
                    }
                }
                offset += strlen(&extended_event_name[offset]);
                offset ++;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }


    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //!!!the order of the internal type_eventtable structure is different for v12.88!!!
        if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
            num_extended = fw_et->event.fw13.num_extended;
            extended_length = fw_et->event.fw13.extended_length;
            extended_event_name = fw_et->event.fw13.extended_event_name;
        }
        else {
            num_extended = fw_et->event.fw.num_extended;
            extended_length = fw_et->event.fw.extended_length;
            extended_event_name = fw_et->event.fw.extended_event_name;
        }

        if ((fw_sysid==456)&&(_appl_version>=0x1400)) {
            //check for "crid=31"
            if (fw_et->crid31_strlen && fw_et->crid31_str) {
//                TAP_Print("\n31:%s", fw_et->crid31_str);
                if (crid31) {
                    if (crid31[0]) {
                        if (strcmp(fw_et->crid31_str, crid31) == 0) {
                            if (repeat_count == 0) {
//                              TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "Repeats", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                TAP_Win_AddItem(&crid_window, "Repeats");
                                dwRow += 20;
                            }
                            repeat_count ++;
//                            sprintf(str1,"%9s", fw_et->crid31_str);
//                            strcpy(str, str1);
                            dump_event(fw_et, j, str1);
                            strcpy(str, str1);
//                          TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                            TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                            TAP_Win_AddItem(&crid_window, str);
                            dwRow += 20;

//                //get the crid for this original entry
//                if (fw_et->crid31_strlen) {
//                    sprintf(str, "\n     %s [S:%s]", fw_et->crid31_str, (fw_et->crid32_strlen)?fw_et->crid32_str:"");
//
////                  TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
//                    TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
//                    TAP_Win_AddItem(&crid_window, str);
//                    dwRow += 20;
//                }


                            if (debug & 1) TAP_Print("\n%s", str);
                        }
                    }
                }
            }
        }
        else {
            offset = 0;
            for (i=0; i<num_extended; i++) {
                ooffset = offset;
                offset += strlen(&extended_event_name[offset]);
                offset ++;

                //check for "crid=31"
                if (strncmp(&extended_event_name[ooffset], "crid=31", 7) == 0) {
                    if (extended_event_name[offset]) {
    //                    TAP_Print("\n31:%s", &extended_event_name[offset]);
                        if (crid31) {
                            if (crid31[0]) {
                                if (strcmp(&extended_event_name[offset], crid31) == 0) {
                                    if (repeat_count == 0) {
//                                      TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "Repeats", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                        TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                        TAP_Win_AddItem(&crid_window, "Repeats");
                                        dwRow += 20;
                                    }
                                    repeat_count ++;
    //                                sprintf(str1,"%9s", &extended_event_name[offset]);
    //                                strcpy(str, str1);
                                    dump_event(fw_et, j, str1);
                                    strcpy(str, str1);
//                                  TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                    TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                    TAP_Win_AddItem(&crid_window, str);
                                    dwRow += 20;
                                    if (debug & 1) TAP_Print("\n%s", str);
                                }
                            }
                        }
                    }
                }
                offset += strlen(&extended_event_name[offset]);
                offset ++;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //!!!the order of the internal type_eventtable structure is different for v12.88!!!
        if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
            num_extended = fw_et->event.fw13.num_extended;
            extended_length = fw_et->event.fw13.extended_length;
            extended_event_name = fw_et->event.fw13.extended_event_name;
        }
        else {
            num_extended = fw_et->event.fw.num_extended;
            extended_length = fw_et->event.fw.extended_length;
            extended_event_name = fw_et->event.fw.extended_event_name;
        }

        if ((fw_sysid==456)&&(_appl_version>=0x1400)) {
            //check for "crid=33"
            if (fw_et->crid33_strlen && fw_et->crid33_str) {
//                TAP_Print("\n33:%s", fw_et->crid33_str);
                if (crid33) {
                    if (crid33[0]) {
                        if (strcmp(fw_et->crid33_str, crid31) == 0) {
                            if (repeat_count == 0) {
//                              TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "Recommendations", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                TAP_Win_AddItem(&crid_window, "Recommendations");
                                dwRow += 20;
                            }
                            recommend_count ++;
//                            sprintf(str1,"%9s", fw_et->crid31_str);
//                            strcpy(str, str1);
                            dump_event(fw_et, j, str1);
                            strcpy(str, str1);
//                          TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                            TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                            TAP_Win_AddItem(&crid_window, str);
                            dwRow += 20;
                            if (debug & 1) TAP_Print("\n%s", str);
                        }
                    }
                }
            }
        }
        else {
            offset = 0;
            for (i=0; i<num_extended; i++) {
                ooffset = offset;
                offset += strlen(&extended_event_name[offset]);
                offset ++;

                //check for "crid=33"
                if (strncmp(&extended_event_name[ooffset], "crid=33", 7) == 0) {
                    if (extended_event_name[offset]) {
    //                    TAP_Print("\n31:%s", &extended_event_name[offset]);
                        if (crid33) {
                            if (crid33[0]) {
                                if (strcmp(&extended_event_name[offset], crid33) == 0) {
                                    if (repeat_count == 0) {
//                                      TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, "Recommendations", COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                        TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                        TAP_Win_AddItem(&crid_window, "Recommendations");
                                        dwRow += 20;
                                    }
                                    recommend_count ++;
    //                                sprintf(str1,"%9s", &extended_event_name[offset]);
    //                                strcpy(str, str1);
                                    dump_event(fw_et, j, str1);
                                    strcpy(str, str1);
//                                  TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                                    TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                                    TAP_Win_AddItem(&crid_window, str);
                                    dwRow += 20;
                                    if (debug & 1) TAP_Print("\n%s", str);
                                }
                            }
                        }
                    }
                }
                offset += strlen(&extended_event_name[offset]);
                offset ++;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

    sprintf(str, "Found %d event%s in series, %d repeat%s, %d recommendations%s", series_count, (series_count==1)?"":"s", repeat_count, (repeat_count==1)?"":"s", recommend_count, (recommend_count==1)?"":"s");
    dwRow = 150;
    dwRow += (1+series_count)*20;
    dwRow += (1+repeat_count)*20;
    dwRow += (1+recommend_count)*20;
//  TAP_Osd_PutS(crid_rgn, dwCol + 10, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1419, FALSE, ALIGN_CENTER);
    TAP_Win_SetFont(&crid_window, ALIGN_CENTER, FNT_Size_1419);
    TAP_Win_AddItem(&crid_window, str);
}



int crid_status()
{
#ifdef EPG_TABLE
    TAP_Print("\nfound %d eit sections", num_tables);
#endif
    return 0;
}

//----------------------------------------------------------------------------
//
//
dword epg_crid(dword param1)
{
    type_eventtable *fw_et;
    int ret;

    if (fw_epg_highlighted_event) {
        fw_et = (type_eventtable *)fw_epg_highlighted_event;
        TAP_ExitNormal();
        ret = dump_crid(fw_et);
    }

    return param1;
}

//----------------------------------------------------------------------------
//
//
int dump_timers()
{
    type_timer_block *tb;
    dword state, subState;
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;
    dword utc, local, now, start, end;
    word start_date;
    word year;
    byte month, day, weekDay;
    word mjd;
    byte hour, min, sec;
    word mjd1;
    byte hour1, min1;

    char event_name[0x100];
    byte num_extended;
    word extended_length;
    byte *extended_event_name;
    int offset, ooffset;

    int found_current = 0;
    char *crid31 = NULL;
    char *crid32 = NULL;
    char *crid33 = NULL;
    int series_count = 0;
    int repeat_count = 0;
    int recommend_count = 0;
    int timers_count = 0;
    TYPE_TapChInfo chInfo;
    char str[0x100];
    char str1[0x100];

    dword dwRgn, dwCol, dwRow, dwWidth, dwHeight, dwMaxCol;


    //delete any old ones
    if (crid_showing) {
        TAP_Win_Delete(&crid_window);
//        TAP_EnterNormal();
//        TAP_ExitNormal();
        if (current_crid == 0) {
            if (fw_epg_refresh) *fw_epg_refresh = 0x18;       //set epg.epg
        }
        crid_showing = 0;
    }

    dwWidth = 500;
    dwCol = ((720 - dwWidth) / 2) - 25;
    dwRow = 100;
    dwHeight = 300;
    dwMaxCol = dwCol + 18 + dwWidth + 18;

    TAP_ExitNormal();
    TAP_Win_SetDefaultColor(&crid_window);
    TAP_Win_Create(&crid_window, crid_rgn, dwCol, dwRow, dwWidth, dwHeight, TRUE, TRUE);
    TAP_Win_SetTitle(&crid_window, "Timers", 1, FNT_Size_1622);
    dwRow = 150;
    TAP_GetState(&state, &subState);
    if (current_crid == 0) {
          if (fw_epg_refresh) *fw_epg_refresh = 0x3c;
    }
    crid_showing = 1;

    sprintf(str1, "\n");
    strcpy(str, str1);



    for(i=0; i<fw_max_timers; i++) {
        tb = (type_timer_block *)((dword)fw_timer_table + (i * fw_timer_table_size));
        if (tb->flags & 0xe0) {
/*
dump((dword)tb, fw_timer_table_size);
            TAP_Print("\ntuner               = %02x", tb->tuner);                  //00  tuner
            TAP_Print("\nflags               = %02x", tb->flags);                  //01, d[7:5] == 2 = valid, d[4:0] = ?
            TAP_Print("\nsatIndex            = %02x", tb->satIndex);               //02  ?
            TAP_Print("\nmode_reservation    = %02x", tb->mode_reservation);       //03  d[7] = svcType, d[6:0] = reservationType
            TAP_Print("\nduration            = %02x", tb->duration);               //04  duration
            TAP_Print("\nsvcNum              = %02x", tb->svcNum);                 //06  svcNum
            TAP_ExtractMjd(tb->startTime>>16, &year, &month, &day, &weekDay);
            TAP_Print("\nstartTime           = (%08x) %02d/%02d/%04d, %02d:%02d", tb->startTime, day, month, year, (tb->startTime>>8)&0xff, (tb->startTime>>0)&0xff);              //08  startTime
            TAP_ExtractMjd(tb->endTime>>16, &year, &month, &day, &weekDay);
            TAP_Print("\nendTime             = (%08x) %02d/%02d/%04d, %02d:%02d", tb->endTime, day, month, year, (tb->endTime>>8)&0xff, (tb->endTime>>0)&0xff);                //0c  endTime
            TAP_Print("\nisRec               = %02x", tb->isRec);                  //10  isRec
            TAP_Print("\nnameFix             = %02x", tb->nameFix);                //11  nameFix setName
            TAP_Print("\nservice_id          = %04x", tb->service_id);             //12  ?                    (10)   service_id ???
            TAP_Print("\nfileName            = %s",   tb->fileName);               //14                       (96=0x60)

            TAP_Print("\ndword54             = %08x", tb->dword54);                //54
            TAP_Print("\nword58              = %04x", tb->word58);                 //58
            TAP_Print("\nword5a [event_id]   = %04x", tb->event_id);               //5a       event_id
            TAP_Print("\nword5c              = %04x", tb->word5c);                 //5c
            TAP_Print("\nword5e [event_id]   = %04x", tb->event_id2);              //5e       event_id
            TAP_Print("\ndword60             = %08x", tb->dword60);                //60
            TAP_Print("\ndword64             = %08x", tb->dword64);                //64
            TAP_Print("\ndword68             = %08x", tb->dword68);                //68
            TAP_Print("\nword6c              = %04x", tb->word6c);                 //6c
            TAP_Print("\nword6e [org_ev_id]  = %04x", tb->event_id3);              //6e       event_id original
            TAP_Print("\ncrid_valid          = %02x", tb->crid_valid);             //70       1=series
            TAP_Print("\nbyte71              = %02x", tb->byte71);                 //71
            TAP_Print("\nbyte72 [2=???]      = %02x", tb->byte72);                 //72       2=
            TAP_Print("\nbyte73              = %02x", tb->byte73);                 //73

//14.06
          TAP_ExtractMjd(tb->crid_date>>16, &year, &month, &day, &weekDay);
          TAP_Print("\ncrid_date           = (%08x) %02d/%02d/%04d, %02d:%02d", tb->crid_date, day, month, year, (tb->crid_date>>8)&0xff, (tb->crid_date>>0)&0xff);              //74       date crid added
          TAP_Print("\ncrid31_str          = %s",   tb->crid31_str);             //78
//14.08
          TAP_Print("\ncrid32_st           = %s",   tb->crid32_str)              //94
          TAP_Print("\ndwordb0             = %08x", tb->dwordb0);                //b0
          TAP_Print("\ndwordb4             = %08x", tb->dwordb4);                //b4

            TAP_Print("\nepg_marker          = %04x", tb->epg_marker);             //74  ?    = 5,            (01)   (02)    2=Rt
            TAP_Print("\nevent_id            = %04x", tb->event_id);               //76
            TAP_Print("\n");
            TAP_Print("\nchannel_number      = %d",   tb->channel_number);         //78                       (0019)         (0016)
            TAP_Print("\nbandwidth           = %02x", tb->bandwidth);              //7a                       (08)           (08)
            TAP_Print("\nunused_7b           = %02x", tb->unused_7b);              //7b                       (00)           (00)
            TAP_Print("\nfrequency           = %d",   tb->frequency);              //7c                       (0007b890)     (00075ad0)
            TAP_Print("\ntransport_stream_id = %04x", tb->transport_stream_id);    //80  ?                    (1004)         (2005)      service_id
            TAP_Print("\nlp_hp_stream        = %02x", tb->lp_hp_stream);           //82
            TAP_Print("\nunused_83           = %02x", tb->unused_83);              //82
            TAP_Print("\noriginal_network_id = %04x", tb->original_network_id);    //84  ?                    (233a)         original_network_id
            TAP_Print("\nnetwork_id          = %04x", tb->network_id);             //86                       (3005)
            TAP_Print("\n");
*/
            timers_count ++;
//          sprintf(str1,"%9s", fw_et->crid32_str);
//          strcpy(str, str1);
            dump_timer(tb, timers_count, str1);

            strcpy(str, str1);
//          TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
            TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
            TAP_Win_AddItem(&crid_window, str);
            dwRow += 20;

            //get the event info for this original entry
            if ((tb->crid_valid == 1) && (tb->event_id3)) {


/*
                fw_et = get_event(tb->service_id, tb->event_id3);
                if (is_firmware(fw_et)) {

                    dump_event(fw_et, 0, str1);
                    sprintf(str, "\n     %04x %04x %04x %s", tb->event_id, tb->event_id2, tb->event_id3, str1);

//                    memcpy(event_name, fw_et->event_name, fw_et->event_name_length);
//                    event_name[fw_et->event_name_length] = '\0';
//                    sprintf(str, "\n     %s", event_name);
                }
                else {
                    sprintf(str, "\n     %04x %04x %04x can't match original event_id", tb->event_id, tb->event_id2, tb->event_id3);
                }
//              TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                TAP_Win_AddItem(&crid_window, str);
                dwRow += 20;
*/


/*
                //get the crid for this original entry
                fw_et = get_event_from_crid(tb->crid31_str);
                if (is_firmware(fw_et)) {

                    dump_event(fw_et, 0, str1);
                    sprintf(str, "\n     %08x %s %s", tb->crid_date, tb->crid31_str, str1);

                }
                else {
                    sprintf(str, "\n     %08x %s", tb->crid_date, tb->crid31_str);
                }
//              TAP_Osd_PutS(crid_rgn, dwCol + 20, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1622, FALSE, ALIGN_LEFT);
                TAP_Win_SetFont(&crid_window, ALIGN_LEFT, FNT_Size_1419);
                TAP_Win_AddItem(&crid_window, str);
                dwRow += 20;
*/
            }


            if (debug & 1) TAP_Print("\n%s", str);

        }
    }


    sprintf(str, "Found %d timer%s", timers_count, (timers_count==1)?"":"s");
    dwRow = 150;
    dwRow += (1+timers_count)*20;
//  TAP_Osd_PutS(crid_rgn, dwCol + 10, dwRow, dwMaxCol, str, COLOR_White, COLOR_None, 0,FNT_Size_1419, FALSE, ALIGN_CENTER);
    TAP_Win_SetFont(&crid_window, ALIGN_CENTER, FNT_Size_1419);
    TAP_Win_AddItem(&crid_window, str);
}
